//-----------------------------------------------------------------------------
// File: CubeMap.cpp
//
// Desc: Example code showing how to do environment cube-mapping.
//
// Copyright (c) 1997-2000 Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
{*********************************************************************
* translated to Delphi: J. Rathlev, Nov. 2001
*           rathlev@physik.uni-kiel.de, J.Rathlev@t-online.de
* required units: D3DApp, D3DFile, D3DFont, D3DUtil, DXTimer from DX8_STUFF
*                 Direct3D8, D3DX8 from Jedi-DirectX8
*   D3App, D3DFile, D3DUtil, D3DX8 and Direct3D8 are revised versions
* required DLL:  D3DX8ab.dll
* required media files:
*   airplane 2.x, lobby_skybox.x, teapot.x, bihull.bmp, wings.bmp,
*   LobbyXPos.bmp, LobbyXNeg.bmp, LobbyYPos.bmp, LobbyYNeg.bmp,
*   LobbyZPos.bmp, LobbyZNeg.bmp from Win-SDk samples
*   set "DefMediaPath" to your local path with these files
*
*********************************************************************}

unit cmmain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Direct3D8, d3dx8, D3DApp, D3DFile, D3DFont, D3DUtil;

const
  DefClientWidth = 640;
  DefClientHeight = 480;
  DefMediaPath = 'media\';
  TrackBallRadius = 0.5;
  D3DFVF_ENVMAPVERTEX = D3DFVF_XYZ or D3DFVF_NORMAL;

// CUBEMAP_RESOLUTION indicates how big to make the cubemap texture.  Larger
// textures will generate a better-looking reflection.
  CUBEMAP_RESOLUTION = 256;

type
//-----------------------------------------------------------------------------
// Name: struct ENVMAPPEDVERTEX
// Desc: D3D vertex type for environment-mapped objects
//-----------------------------------------------------------------------------
  TENVMAPPEDVERTEX = record
    p,n : TD3DXVECTOR3; // Position, Normal
    end;


  TCubeMapApp = class(TD3DApplication)
  private
    Font        : TD3DFont;           // Font for drawing text
    ShinyTeapot : TD3DMesh;
    SkyBox      : TD3DMesh;
    Airplane    : TD3DMesh;
    matAirplane : TD3DXMATRIX;
    CubeMap     : IDIRECT3DCUBETEXTURE8;
    CubeFaceZBuffer : IDirect3DSurface8;
    MediaPath : string;

    constructor Create (AOwner : TComponent; AMediapath : string);
    procedure RenderScene(bRenderTeapot : boolean);
    function RenderSceneIntoCubeMap : HRESULT;
    function ConfirmDevice(pCaps : TD3DCAPS8;
           dwBehavior : LongWord ; Format : TD3DFORMAT) : HResult; override;
    function InitDeviceObjects : HRESULT; override;
    function RestoreDeviceObjects : HRESULT; override;
    function FrameMove : HRESULT; override;
    function Render : HRESULT; override;
    function InvalidateDeviceObjects : HRESULT; override;
    function DeleteDeviceObjects : HRESULT; override;
    end;

  TForm1 = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private-Deklarationen }
  public
    { Public-Deklarationen }
    CubeMapApp : TCubeMapApp;
    MediaPath : string;
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

constructor TCubeMapApp.Create (AOwner : TComponent; AMediapath : string);
var
  i : integer;
begin
  inherited Create (AOwner);
  strWindowTitle:='CubeMap: Environment cube-mapping';
  bUseDepthBuffer:= TRUE;
  Font:=TD3DFont.Create('Arial',12,D3DFONT_BOLD);
  SkyBox:=TD3DMesh.D3DMesh('SkyBox',AMediapath);
  ShinyTeapot:=TD3DMesh.D3DMesh('Teapot',AMediapath);
  Airplane:=TD3DMesh.D3DMesh('Airplane',AMediapath);
  CubeMap:=nil;
  CubeFaceZBuffer:=nil;
  end;

//-----------------------------------------------------------------------------
// Name: RenderScene()
// Desc: Renders all visual elements in the scene. This is called by the main
//       Render() function, and also by the RenderIntoCubeMap() function.
//-----------------------------------------------------------------------------
procedure TCubeMapApp.RenderScene (bRenderTeapot : boolean);
var
  matWorld,
  matViewTmp, matProjTmp,
  matProjSave, matViewSave : TD3DXMATRIX;
  hr : HResult;
  s : string;
begin
  // Clear the viewport
  with d3dDevice do begin
    Clear( 0, nil ,D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,0), 1.0, 0);

    // Set the texture stage states
    SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
    SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1 );
    SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
    SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
    SetTextureStageState( 0, D3DTSS_ADDRESSU,  D3DTADDRESS_MIRROR );
    SetTextureStageState( 0, D3DTSS_ADDRESSV,  D3DTADDRESS_MIRROR );

    // Render the Skybox
    // Save the current matrix set
    GetTransform( D3DTS_VIEW,      matViewSave);
    GetTransform( D3DTS_PROJECTION,matProjSave);

    // Disable zbuffer, center view matrix, and set FOV to 90 degrees
    matViewTmp:=matViewSave;
    matProjTmp:=matViewSave;
    with matViewTmp do begin
      _41:=0;  _42:=0; _43:=0;
      end;
    D3DXMatrixPerspectiveFovLH(matProjTmp, D3DX_PI/2, 1.0, 0.5, 10000.0 );
    SetTransform( D3DTS_PROJECTION,matProjTmp );
    SetTransform( D3DTS_VIEW,matViewTmp );
    SetRenderState( D3DRS_ZENABLE,D3DZB_FALSE);

    // Render the skybox
    D3DXMatrixIdentity(matWorld );
    SetTransform( D3DTS_WORLD,matWorld );
    SkyBox.Render( d3dDevice,true,true);

    // Restore the render states
    SetTransform( D3DTS_VIEW,      matViewSave);
    SetTransform( D3DTS_PROJECTION,matProjSave);
    SetRenderState( D3DRS_ZENABLE,1);

    // Render the main file-based object
    SetTransform( D3DTS_WORLD,matAirplane );
    Airplane.Render( d3dDevice,true,true);

    // Render the object with the environment-mapped body
    if bRenderTeapot then begin
      D3DXMatrixIdentity(matWorld );
      SetTransform( D3DTS_WORLD,matWorld );

      // Turn on texture-coord generation for cubemapping
      SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR );
      SetTextureStageState( 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT3 );

      // Render the object with the environment-mapped body
      SetTexture( 0, CubeMap );
      ShinyTeapot.Render( d3dDevice,true,true);

      // Restore the render states
      SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU );
      SetTextureStageState( 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE );
      end;
    end;
  end;

//-----------------------------------------------------------------------------
// Name: RenderSceneIntoCubeMap()
// Desc: Renders the scene to each of the 6 faces of the cube map
//-----------------------------------------------------------------------------
function TCubeMapApp.RenderSceneIntoCubeMap : HRESULT;
var
  matViewDir, matViewTmp, matProjTmp,
  matProjSave, matViewSave : TD3DXMATRIX;
  pBackBuffer, pZBuffer : IDIRECT3DSURFACE8;
  pCubeMapFace : IDIRECT3DSURFACE8;
  i : integer;
begin
  // Save transformation matrices of the device
  with d3dDevice do begin
    GetTransform(D3DTS_VIEW,matViewSave);
    GetTransform( D3DTS_PROJECTION,matProjSave);
    // Set the projection matrix for a field of view of 90 degrees
    D3DXMatrixPerspectiveFovLH(matProjTmp, D3DX_PI/2, 1.0, 0.5, 100.0);
    SetTransform( D3DTS_PROJECTION,matProjTmp);

    // Get the current view matrix, to concat it with the cubemap view vectors
    GetTransform( D3DTS_VIEW,matViewDir );
    with matViewDir do begin
      _41:= 0.0; _42:= 0.0; _43:= 0.0;
      end;
    // Store the current backbuffer and zbuffer
    GetRenderTarget(pBackBuffer );
    GetDepthStencilSurface(pZBuffer );

  // Render to the six faces of the cube map
    for i:=0 to 5 do begin
      // Set the view transform for this cubemap surface
      matViewTmp:=D3DUtil_GetCubeMapViewMatrix( TD3DCUBEMAP_FACES(i) );
      D3DXMatrixMultiply(matViewTmp,matViewDir,matViewTmp );
      SetTransform(D3DTS_VIEW,matViewTmp );

      // Set the rendertarget to the i'th cubemap surface
      CubeMap.GetCubeMapSurface( TD3DCUBEMAP_FACES(i), 0, pCubeMapFace );
      SetRenderTarget( pCubeMapFace, CubeFaceZBuffer);

      // Render the scene (except for the teapot)
      BeginScene;
      RenderScene(FALSE );
      EndScene;
      end;

    // Change the rendertarget back to the main backbuffer
    SetRenderTarget( pBackBuffer, pZBuffer );

    // Restore the original transformation matrices
    SetTransform( D3DTS_VIEW,matViewSave);
    SetTransform( D3DTS_PROJECTION, matProjSave);
    end;

  result:=S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function TCubeMapApp.ConfirmDevice(pCaps : TD3DCAPS8;
                              dwBehavior : longword;
                              Format : TD3DFORMAT) : HResult;

begin
  Result:=S_OK;
  // GetTransform doesn't work on PUREDEVICE
  if(dwBehavior and D3DCREATE_PUREDEVICE <>0) then result:=E_FAIL
  else begin
    // This sample uses alpha textures and/or straight alpha. Make sure the
    // device supports them
    if( pCaps.TextureCaps and D3DPTEXTURECAPS_CUBEMAP<>0) then exit;
    // Check that we can create a cube texture that we can render into
    if (FAILED(D3D.CheckDeviceFormat(pCaps.AdapterOrdinal,
        pCaps.DeviceType, Format, D3DUSAGE_RENDERTARGET,
        D3DRTYPE_CUBETEXTURE, Format))) then result:=E_FAIL;
    end;
  end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function TCubeMapApp.FrameMove : HRESULT;
var
  matTrackBall,
  matTrans,
  matView,
  mat1       : TD3DXMATRIX;
  quat       : TD3DXQUATERNION;
begin
  result:=E_FAIL;
  // Animate file object
  D3DXMatrixScaling(matAirplane, 0.2, 0.2, 0.2 );
  D3DXMatrixTranslation(mat1, 0.0, 2.0, 0.0 );
  D3DXMatrixMultiply(matAirplane,matAirplane,mat1 );
  D3DXMatrixRotationX(mat1,-2.9*dwTime/1000 );
  D3DXMatrixMultiply(matAirplane,matAirplane,mat1 );
  D3DXMatrixRotationY(mat1,1.055*dwTime/1000 );
  D3DXMatrixMultiply(matAirplane,matAirplane,mat1 );

  // When the window has focus, let the mouse adjust the camera view
  if GetFocus<>0 then begin
    quat:=D3DUtil_GetRotationFromCursor(hWnd,TrackBallRadius);
    D3DXMatrixRotationQuaternion(matTrackBall,quat );
    D3DXMatrixTranslation(matTrans, 0.0, 0.0, 5.0 );
    D3DXMatrixMultiply(matView,matTrackBall,matTrans );
    d3dDevice.SetTransform(D3DTS_VIEW,matView );
    end;
  // Render the scene into the surfaces of the cubemap
  if FAILED(RenderSceneIntoCubeMap) then exit;
  result:=S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function TCubeMapApp.Render : HRESULT;
begin
  // Begin the scene
  if SUCCEEDED( d3dDevice.BeginScene) then begin
    // Render the scene, including the teapot
    RenderScene( TRUE );
    // Output statistics
    with Font do begin
      DrawText( 2,  0, D3DCOLOR_ARGB(255,255,255,0),pchar(strFrameStats),0);
      DrawText( 2, 20, D3DCOLOR_ARGB(255,255,255,0),pchar(strDeviceStats),0);
      end;
    // End the scene.
    d3dDevice.EndScene();
    end;
  Result := S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
function TCubeMapApp.InitDeviceObjects : HRESULT;
begin
  // Load the file objects
  // Initialize the font's internal textures
  Result:=Font.InitDeviceObjects( d3dDevice );
  if Failed(Result) then Exit;
  if( FAILED( ShinyTeapot.Create1( d3dDevice, 'teapot.x') ) ) then begin
    result:=D3DAPPERR_MEDIANOTFOUND; exit;
    end;
  if( FAILED( SkyBox.Create1( d3dDevice,'lobby_skybox.x') ) ) then begin
    result:=D3DAPPERR_MEDIANOTFOUND; exit;
    end;
  if( FAILED( Airplane.Create1( d3dDevice,'airplane 2.x') ) ) then begin
    result:=D3DAPPERR_MEDIANOTFOUND; exit;
    end;

  // Set mesh properties
  Airplane.SetFVF( d3dDevice, D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1 );
  ShinyTeapot.SetFVF( d3dDevice, D3DFVF_ENVMAPVERTEX );
  ShinyTeapot.UseMeshMaterials( FALSE );
  Result := S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function TCubeMapApp.RestoreDeviceObjects : HRESULT;
var
  fAspect : single;
  mtrl  : TD3DMATERIAL8;
  light : TD3DLIGHT8;
  matProj : TD3DXMatrix;
begin
  Result:=E_Fail;
  // InitDeviceObjects for file objects (build textures and vertex buffers)
  ShinyTeapot.RestoreDeviceObjects( d3dDevice );
  SkyBox.RestoreDeviceObjects( d3dDevice );
  Airplane.RestoreDeviceObjects( d3dDevice );
  Font.RestoreDeviceObjects;

  // Create the cubemap, with a format that matches the backbuffer, since
  // we'll be rendering into it
  if FAILED(d3dDevice.CreateCubeTexture(CUBEMAP_RESOLUTION, 1, D3DUSAGE_RENDERTARGET,
                                           d3dsdBackBuffer.Format,
                                           D3DPOOL_DEFAULT,CubeMap)) then exit;

  // We create a separate Z buffer for the cube faces, because user could
  // resize rendering window so that it is smaller than a cube face. In
  // this case we cannot use the rendering window Z buffer for cube faces.
  if FAILED( d3dDevice.CreateDepthStencilSurface(
              CUBEMAP_RESOLUTION, CUBEMAP_RESOLUTION,
              D3DFMT_D16, D3DMULTISAMPLE_NONE,CubeFaceZBuffer))  then exit;
  // Set default render states
  with d3dDevice do begin
    SetRenderState( D3DRS_ZENABLE,  1);
    SetRenderState( D3DRS_AMBIENT,  $00aaaaaa );
    SetRenderState( D3DRS_LIGHTING, 1);
    end;
  // Set the transform matrices
  with d3dsdBackBuffer do fAspect:= Width/Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 0.5, 100.0);
  with d3dDevice do begin
    SetTransform( D3DTS_PROJECTION,matProj );
    // Setup a material
    D3DUtil_InitMaterial( mtrl, 1.0, 1.0, 1.0, 1.0 );
    SetMaterial( mtrl );
    end;
  // Set up a light
  if( d3dCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS <>0) then begin
    D3DUtil_InitLight( light, D3DLIGHT_DIRECTIONAL, 0.0, 0.0, 1.0 );
    with light.Ambient do begin
      r:= 0.3; g:= 0.3; b:= 0.3;
      end;
    with d3dDevice do begin
      SetLight( 0, light );
      LightEnable( 0, TRUE );
      SetRenderState( D3DRS_LIGHTING, 1);
      end;
    end;
  Result := S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function TCubeMapApp.InvalidateDeviceObjects : HRESULT;
begin
  ShinyTeapot.InvalidateDeviceObjects;
  SkyBox.InvalidateDeviceObjects;
  Airplane.InvalidateDeviceObjects;
  Font.InvalidateDeviceObjects;
  Result := S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function TCubeMapApp.DeleteDeviceObjects : HRESULT;
begin
  Font.DeleteDeviceObjects();
  ShinyTeapot.Destroy();
  SkyBox.Destroy();
  Airplane.Destroy();
  Result := S_OK;
  end;

{ ------------------------------------------------------------------- }
procedure TForm1.FormCreate(Sender: TObject);
begin
  ClientWidth:=DefClientWidth; ClientHeight:=DefClientHeight;
//  MediaPath:='f:\win-sdk\samples\multimedia\media\';
  MediaPath:=DefMediaPath;
  CubeMapApp:=TCubeMapApp.Create(self,MediaPath);
  Caption:=CubeMapApp.strWindowTitle;
  end;

procedure TForm1.FormShow(Sender: TObject);
begin
  // 3D Initialization
  if Failed(CubeMapApp.Init) then Close;
  end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if assigned(CubeMapApp) then CubeMapApp.Cleanup3DEnvironment;
  end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_ESCAPE then close;
  end;

end.
